"""
data是classicML中关于数据操作的模块,
它可以实现数据的自动加载和预处理等功能,
并开放大量的预处理函数的接口.
"""
from classicML.backend.python.data import preprocessing
from classicML.backend.python.data.dataset import Dataset

__version__ = 'backend.python.data.0.2'
