//
// dtypes.h
// dtypes
// Created by 孙瑞琦 on 2021/7/19.
//
//

#ifndef CLASSICML_BACKEND_CC_DTYPES_H
#define CLASSICML_BACKEND_CC_DTYPES_H

typedef float float32;
typedef double float64;

typedef std::int32_t int32;
typedef std::int64_t int64;

typedef std::uint8_t uint8;
typedef std::uint32_t uint32;

#endif /* CLASSICML_BACKEND_CC_DTYPES_H */